library(methods)

# Description of a SOAP type including a function
# that can convert to and from a SOAP representation.
setClass("SOAPType", representation(name="character",
                                    ns="character",
                                    nsuri="character"))

setClass("SOAPEnvelope",
          representation())

setClass("SOAPActor", representation())


# Representations of different error types that
# can occur in sending a SOAP message.
setClass("SOAPFault",
          representation(message="character",
                         actor="SOAPActor",
                         detail="ANY"))

setClass("SOAPVersionMismatchFault", representation("SOAPFault"))

setClass("SOAPMustUnderstandFault", representation("SOAPFault"))

setClass("SOAPClientFault", representation("SOAPFault"))
setClass("SOAPServerFault", representation("SOAPFault"))

setClass("SOAPGeneralFault", representation("SOAPFault", code="character"))



# Descriptions of SOAP Servers that allow one to connect
# to them.  
setClass("SOAPServer", representation(
                                      host = "character",
                                      port = "integer",
                                      url = "character"),
                       prototype = list(host = "", port = as.integer(NA), url="" ))


setClass("DynamicSOAPServer", representation(server = "SOAPServer", functions = "list")) # , contains = "SOAPServer")




setClass("HTTPSOAPServer", representation("SOAPServer"))
setClass("HTTPSSOAPServer", representation("HTTPSOAPServer"))

setClass("FTPSOAPServer", representation("SOAPServer"),
           prototype = list(host = "", port = as.integer(NA), url="" ))


# A simple class to represent a complete SOAP action string
# so that we can distinguish it from a regular string.
# Such objects are assumed to be the complete action name
# whereas a regular string will be combined with the
# method name to create the action value.
setClass("SOAPAction", representation("character"))

# A description of a SOAP method in a server.
# See http://www-128.ibm.com/developerworks/webservices/library/ws-whichwsdl/ for
# information about bindingStyle and use fields.
setClass("WSDLMethod", representation(name="character",
                                      parameters = "list",
                                      returnValue="SOAPType",
                                      action="SOAPAction",
                                      namespace="character",
                                      bindingStyle = "character", # rpc or document.
                                      use = "character",          # literal or encoded
                                      documentation = "character"
                                     ),
                       prototype = list(bindingStyle = as.character(NA),
                                        use = c(input = as.character(NA), output = as.character(NA)),
                                        documentation = as.character(NA)))


setClass("SOAPVoidType", representation("SOAPType"),
            prototype = list(name = "null", ns = "xsi"))
setClass("SOAPAnyType", representation("SOAPType"))

setClass("SOAPTypeReference", representation("SOAPType"))

setClass("TerminalSOAPType", representation("VIRTUAL"))

setClass("BasicSOAPType", representation("SOAPType",
                                         fromConverter = "function",
                                         toConverter = "function"))


#XXX Is this a setClassUnion ?
setClass("PrimitiveSOAPType", contains=c("BasicSOAPType", "TerminalSOAPType"))



# Meta information (typically) read from a WSDL file that defines
# the properties of a SOAP server. This includes how to find the server
# the methods and any new types it defines for its purposes.
#
setClass("SOAPServerDescription", representation(name = "character",
                                      server = "SOAPServer",
                                      operations = "list",
                                      types = "list"
                                     ))


setClass("ArrayType", representation("BasicSOAPType", elementType = "character", elType = "SOAPType", dims = "integer"))

# A definition for a complex data type defined in a WSDL file.
setClass("ClassDefinition", representation("BasicSOAPType",  slotTypes = "list"), contains = "TerminalSOAPType")

setClass("EnumValuesDef",  contains = "character")

setClass("ExtendedClassDefinition", representation("ClassDefinition", base = "character"))

# Stores the output from a SOAP request, i.e. the resulting HTTP message
# This contains both the HTTP hearder info and the payload.
setClass("SOAPResult", representation(header="list", content = "character"))

#
# When we generate an S4 class to represent a SOAP compound data type,
# we have it extend this degenerate class. This allows us to
# provide methods that dispatch on that class such as converting from
# S to SOAP values and vice-verse.
setClass("VirtualSOAPClass")
