verbose = FALSE

setClass("SOAPClientInterface",
           representation(functions = "list",
                          classes = "list"))



capitalizeFirstLetter =
function(word)
{
  els = strsplit(word, "")[[1]]
  els[1] = toupper(els[1])
  paste(els, collapse="")
}

convertToSName =
function(name, useDash = ifelse("UseDashInSOAPNames" %in% names(options()), getOption("UseDashInSOAPNames"), TRUE))
{
  if(useDash)
    return(name)
  
  els = strsplit(name, "_")[[1]]
  if(length(els) == 1)
    return(els)
  
  els[2:length(els)] = sapply(els[-1], capitalizeFirstLetter)

  paste(els, collapse="")
}

dynamicSOAPServer =
function(iface, obj = new("DynamicSOAPServer"))
{
  server = environment(iface@functions[[1]])$".defaultServer"
  obj@server = server
  obj@functions = iface@functions
  
  obj
}


setMethod("$", c("DynamicSOAPServer", "character"),
           function(x, name) {
               f = x@functions[[name]]
                # Need to put this server object as the default server.
                # formals(f)
               f
           })
setGeneric("names", function(x) standardGeneric("names"))

setMethod("names", "DynamicSOAPServer",
  function(x)
    names(x@functions)
)

genSOAPClientInterface =
  #
  # def is from reading the WSDL file
  # name is the prefix for the methods
  # env is where to put the methods.
  #
function(operations = def@operations[[1]], def, name = def@name, env = new.env(), where = 1, server = def@server, verbose = FALSE)  
{

 if(!is.null(where)) {
  classes = list()
  for(i in def@types) {

   if(verbose)
       cat("Attempting to define class for", i$name,"\n")    
   if(is(i$definition, "ClassDefinition")) {
     classes[[i$name]] = createTypeClass(i, def@types, where = where)
   } else if(is(i$definition, "ArrayType")) {
     classes[[i$name]] = createArrayClass(i, def@types, where = where)
   }
 }
 } else
   classes = NULL


 funs = list()
 for(i in operations) {
   if(verbose)
     cat("Operation", i@name, "\n")

   funs[[convertToSName(i@name)]] = createOperationDefinition(i, server, def@types, env)
 }

 
 ans = new("SOAPClientInterface", functions = funs, classes =classes)

 ans
}


getXSDSchemaURIs =
function() {
  sapply(.SOAPDefaultNameSpaces, "[", "xsd")
}  

coerceArgumentCode =
function(id, type)
{
  name = convertToSName(id)

  if(length(type@nsuri) && type@nsuri %in% getXSDSchemaURIs()) {
     which = match(type@name, sapply(SOAPTypes, "[[", "type"))
     ifelse(!is.na(which),
             paste("as.",  names(SOAPTypes)[which], "(", id, ")", sep = ""),
            id)
  } else {
     paste("as(", id, ", '", type@name, "')", sep = "")
  }
}  


createOperationDefinition =
function(.operation, .defaultServer, .types, env = globalenv())
{
   # resolve the parameter types
 .operation@parameters = lapply(.operation@parameters, resolve, .types)

 .operation@returnValue = rval = resolve(.operation@returnValue, .types)

   # Strip the namespace prefix.
 returnClass <- ifelse(rval@name %in% gsub(".*:", "", names(.types)), rval@name, NA)
 fixes = character(2)
 if(!is.na(returnClass)) {
    fixes[1] = "  ans <-\n"
    fixes[2] = paste("\tas(ans, '",  returnClass, "')\n", sep = "")
 }
 
 
 if(.operation@action == "")
     # NULL is not used, so just NULL for placeholder.
   .operation@action[1] = .SOAPAction(.operation@action, .operation@name, NULL, .operation@namespace)

 dotArgs = ""
 if(length(.operation@parameters)) {
   values = sapply(names(.operation@parameters),  function(x) coerceArgumentCode(x, .operation@parameters[[x]]))
   dotArgs = paste(paste(paste("'", names(.operation@parameters), "'", sep=""),
                                       values, sep = " = "),
                   collapse=", ")
 }

 
 txt = paste("function(",
              paste(c(sapply(names(.operation@parameters), convertToSName),
                      "server = .defaultServer, .convert = .operation@returnValue, .opts = list(), ..."),
                    collapse=",\n\t "),
             ")\n{\n",
              fixes[1],
             "\t .SOAP(server, .operation@name, ",
                  "\n\t\t",         dotArgs,
                   ifelse(length(.operation@parameters), ", ", ""),
                  "\n\t\t",         "action = .operation@action, ",
                  "\n\t\t",         "xmlns = .operation@namespace, ",
                  "\n\t\t",         ".types = .operation@parameters, ",
                  "\n\t\t",         ".convert = .convert,",
                  "\n\t\t",         ".header = .header,",
                  "\n\t\t",         ".opts = .opts ",
                  ifelse(.operation@use["input"] == "literal", ",\n\t\t .literal = TRUE", ""),
                  ")\n",
                  fixes[2],
                  "\n}", sep="")  
 f = eval(parse(text = txt)) 

 environment(f) <- e <- new.env()
 e$.operation <- .operation
 e$.header <- getSOAPRequestHeader(.operation@action, .server = .defaultServer)
 if(!is.null(.defaultServer))
    e$.defaultServer <- .defaultServer
 
 parent.env(e) = env
 
 attr(f, "source") <- txt
 
 rm(txt)
 rm(dotArgs)
 
 f
}


setGeneric("convertFromSOAP", function(val, type) standardGeneric("convertFromSOAP"))

setMethod("convertFromSOAP", "SOAPResult", 
          function(val, type) {
            node <- parseSOAP(val@content, asText = TRUE)
            node <- getReturnNode(node)
            
            fromSOAP(node, type = type)
          })



createTypeClass =
function(type, types, where = globalenv(), parentClass = "VirtualSOAPClass",
         namespaceDefs = list())
{
  if(verbose)
    cat("[createTypeClass]", type$name, "\n")
  repn = createClassRepresentation(type$definition, types, namespaceDefs)
  setClass(type$name, representation = repn, where = where, contains = parentClass)
}


createArrayClass =
  # Should be merged into createTypeClass.sl
function(type, types, where = globalenv(), parentClass = "VirtualSOAPClass")
{
  name = type$definition@elementType
  el = type$definition@elType

  if(el@nsuri %in% getXSDSchemaURIs()) {
    builtinClass = getArrayElementTypeFromName(name)
    if(is.na(builtinClass))
      stop("can't identify array element type for SOAP definition")

    setClass(type$name, contains = builtinClass, where = where)    
  } else {
        #XXX want a typed list where the elements are checked. See DCOM code.
     setClass(type$name, contains = "list", where = where)    
  }
}


getArrayElementTypeFromName =
function(name, stripArray = TRUE, convertToRType = TRUE)
{
 
 els = strsplit(name, ":")[[1]]
 if(length(els) > 1) {
   els = els[2]
 }
 if(stripArray)
   els = gsub("\\\[\\\]$", "", els)

 if(!convertToRType)
   return(els)

 i = match(paste("xsd:", els, sep = ""), sapply(SOAPTypes, function(x) x[["xsi:type"]]))
 if(!is.na(i))
   els = names(SOAPTypes)[i]
 
 els
}  


newSOAPClass =
  #
  # Creates a new instance of the specified class (className) and populates its
  # fields with the values from the XML node.
  #
  #XXX converters is not used here yet.
function(node, className, converters = SOAPPrimitiveConverters, type = NULL)
{
 obj = new(className)
 
 for(i in slotNames(className)) {
   tmp = node[[i]]
   if(!is.null(type))
     slotType = type@slotTypes[[i]]
   else
     slotType = NULL

   slot(obj, i) <- fromSOAP(tmp, type = slotType, findReturn = FALSE) #XXX Need SOAP type here!
 }

 
  obj
}


createClassRepresentation =
 #
 # 
function(type, types, namespaceDefs = list())
{
   repn = lapply(type@slotTypes, mapSOAPTypeToS, types, namespaceDefs)
   nas = sapply(repn, is.na)
   if(any(nas))
     stop("problem resolving SOAP type ", names(type)[nas])

   repn
}

mapSOAPTypeToS =
  #
  # Take a SOAP type (by name or as a SOAPType object)
  # and find the name of the R data type that it maps to 
  # so that data type can be used in, e.g., a class representation
  # e.g. "int" goes to "integer".
  #
  #XXX Need to deal with namespaces. Tricky because they may have been
  # defined elsehwere, not in the top of the document. Need to match
  # the prefix to the actual URI and then compare URIs.
function(type, types = list(), namespaceDefs = list())
{
    if(is(type, "SOAPType"))
      type = type@name

    if(length(type) == 1) {
      type =  strsplit(type, ":")[[1]]
      if(length(type) == 2)
        return(mapSOAPTypeToS(type, types, namespaceDefs))
    }

    if(length(type) == 2) {
      ns = type[1]
      type = type[2]
    }

    which = sapply(SOAPTypes, function(x) x[["type"]] == type) #is this now correct?
    if(any(which))
      return((names(SOAPTypes)[which])[1])

#    if(length(grep(":", type)))
#      type = gsub("^.*:", "", type)
    
    if(length(types)) {
      which = match(type, sapply(names(types), discardNamespace))
      if(!is.na(which))
        return(discardNamespace(names(types)[which]))
    }

    stop("Can't match SOAP type ", type)
}



if(FALSE) {

# f = function(){}
# formals(f) <- alist(server=, kid=, threshold=, orgs=)
  
 body(f) <- substitute({
  val = .SOAP(server,
              .opName,
              action = .action,
              xmlns = .namespace,
              .types = ..types)
  }, 
  list(.opName= operation@name,
       .action = operation@action,
       .namespace = operation@namespace,
       ..types = operation@parameters))

  # now put the arguments in.
 e = body(f)[[2]]
 kk = e[4:length(e)]
# for(i in )
 k = body(f)[[2]][[3]]
}

# body(f) = substitute(body(f), list(.opname=operation@name))
