SOAPAction =
  #
  # Constructor for a SOAPAction object typically
  #  read from a WSDL file.
function(x, obj = new("SOAPAction"))
{
  as(obj, "character") = x

  obj
}


SOAPType =
  #
  # 
  #
function(name, ns = "", nsuri="", namespaceDefs = list(), obj = new("SOAPType"))
{
  if(is.null(name)) 
    return(new("SOAPVoidType"))


   # If the name is in the form of namespace:type, split it
   # and assign to name and ns separately.
   # e.g. xsd:int will be come name = "int", ns = "xsd"
  name = strsplit(name, ":")[[1]]
  
  if(length(name) == 2) {
    if(missing(ns))
       ns = name[1]
    name = name[2]        
  }


  # resolve namespace.
  if(missing(nsuri) && length(namespaceDefs)) {
     idx = match(ns, names(namespaceDefs))
     if(is.na(idx) && any(names(namespaceDefs) == ""))
        idx = which(names(namespaceDefs) == "")

     if(!is.na(idx)) {
         nsuri = namespaceDefs[[idx]]$uri
     }
  }

  
  if(nsuri %in% getXSDSchemaURIs()) {
    # tmp = paste(ns, name, sep=":")
    which = sapply(SOAPTypes, function(x) name == x[["type"]])
    if(any(which)) {
       obj = new("PrimitiveSOAPType")
       obj@toConverter = get(paste("as", names(SOAPTypes)[which], sep="."))
       tmp = SOAPTypes[which][[1]]
       if("from" %in% names(tmp)) {

           if(is.character(tmp$from)) {
              if(length(tmp$from) > 1) {
                tmp$from = get(tmp$from[1], paste("package", tmp$from[2], sep = ":"))
              } else
                tmp$from = get(tmp$from)   
           }          
           obj@fromConverter = tmp$from
       }
    }
  } else {
     # Should check that this is referring to the target namespace of the
     # document, e.g. targetNamespace and xmlns:[targetNamespace] in the root node
     obj = new("SOAPTypeReference")
  }

  obj@name = name
  obj@ns = ns
  obj@nsuri = nsuri

  obj
}

SOAPServerDescription =
function(name,  server, operations, types, obj = new("SOAPServerDescription"))
{
 obj@name = name
 obj@server = server
 if(!is.null(operations))
   obj@operations = operations

 if(!is.null(types))
   obj@types = types

 obj
}


WSDLParseHandlers =
function(baseURI, verbose = FALSE)
{
  baseURI = dirname(baseURI)

  list(comment = function(x) NULL,
       import = function(node) {
          location = xmlGetAttr(node, "location", NA)
          if(is.na(location))
            return(node)
          
             #XXX Merge the location with the base taking care of absolute addresses, etc.
             # Is this in the XML package?
          location = paste(baseURI, location, sep = "/")
          if(verbose)
             cat("Importing", location, "as part of", baseURI, "\n")
          xmlRoot(xmlTreeParse(location))
       })
}  


processWSDL <-
  #
  # Currently only handle one <service> element
  #
function(fileName = "KEGG.wsdl", handlers = WSDLParseHandlers(fileName))
{
 library(XML)
  # Parse the WSDL file into a tree which we will pick apart.
  # Could put handlers to convert the elements on the fly.
 doc = xmlRoot(xmlTreeParse(fileName, handlers = handlers,  asTree = TRUE))

  # Get the (first) server declared in <service><port>...
 port = doc[["service"]][["port"]]

 loc = xmlGetAttr(port[["address"]], "location")

 server = SOAPServer(loc)

 types = processWSDLTypes(doc[["types"]], doc) 
 
  # Now find the methods/functions that are available from this server.
  # These are in the <binding> elements. We should match each one to
  # each <service>. Currently only handle one <service> element
 
 tmp = doc[names(doc) == "binding"]
 ops = lapply(tmp, processWSDLBindings, doc, types)
 names(ops) = sapply(tmp, function(x) xmlGetAttr(x, "name"))


 SOAPServerDescription(name = xmlGetAttr(port, "name"), server = server,
                       operations = ops, types = types)
}

processWSDLTypes =
   # Process all the elements in types (hopefully <schema> elements)
   # and merge. We may want to keep these as a list indexed by namespace/schema identifier.
   # Originally, we just processed the first one.
   # types = processWSDLTypes(doc[["types"]][[1]], doc)   
function(node, doc)
{
  if(is.null(node))
      return(list())

   # Find the URIs of the namespaces for the different schema
 ns = xmlSApply(node, function(x) xmlGetAttr(x, "targetNamespace"))
   # map these to namespace identifiers/short-names
 nsDefs = doc$namespaceDefinitions
 which = match(ns, sapply(nsDefs, function(x) x$uri))
 ns = names(nsDefs)[which]

 types = lapply(1:xmlSize(doc[["types"]]),
                function(i) {
                   processSchemaTypes(doc[["types"]][[i]], doc, ns = ns[i], nsDefs)
                 })
  # Fix the names on these types to avoid the schema.
 types = unlist(types, recursive = FALSE)
 names(types) = gsub("^schema\.", "", names(types))

 types
}


processSchemaTypes =
  # XXX need to handle simple types and simple content.  
function(types,  doc, ns = character(0), namespaceDefs = list())
{
 if(is.null(types))
   return(NULL)

 tmp = xmlApply(types, function(x, ...) {
                           name = ""
                           if(xmlName(x) == "element") {
                               # need the name from here.
                             name = xmlGetAttr(x, "name")

                             if(xmlSize(x)) {
                                 # Now get the content inside the element.
                                x = x[[1]]
                                x$attributes <- c(x$attributes, name = name)
                             } else {
                              tp = xmlGetAttr(x, "type")
                              #XXX now resolve this from the earlier elements
                             }
                           }

                            # XXX
                           if(xmlName(x) %in% c("complexType", "complexContent")) {
                                processWSDLType(x, ...)
                           } else
                                NULL
                           },
                         types, namespaceDefs = namespaceDefs)
 names(tmp) = sapply(tmp, function(x)
                              paste(ns, x$name, sep = ":"))

  # Drop the NULL elements which correspond to non-complexType elements,
  # e.g. <import namespace="..">
 tmp = tmp[!sapply(tmp, is.null)]

 tmp  
}



# This material relates to XML schema and processing of types defined within
# schema. Ideally, we will separate this into a separate package and use the
# results in different ways, not just for SOAP. And we will provide a more
# extensive framework for dealing with all of the schema details rather than
# the rather limited but common ones supported here.
#
#  See the O'Reilly book XML Schemas by Eric van der Vlist for more information
#  on schemas, or look on the Web (e.g. www.xml.com/schemas)
#

processWSDLType =
  #
  # This currently deals with a very small subset of XML schema
  # specifications.  These are the common ones, but more work 
  # is definitely needed.
  #
function(type, types, namespaceDefs = list())
{
# Rationalize this 

  name = xmlGetAttr(type, "name", "")

  if(name == "simpleType") {
    if(xmlSize(type) > 0 && xmlName(type[[1]]) == "restriction") {
         # Assume enumeration values for present.
       def <- new("EnumValuesDef", xmlSApply(type[[1]],  xmlGetAttr, "value"))
    } else
      def <- "xsd:string"
    return(def)
  }

  if(xmlName(type) == "complexContent")
    tmp = type
  else if(xmlName(type) == "complexType" && xmlName(type[[1]]) == "all")
    tmp = type
  else
    tmp <- type[["complexContent"]]

  if(!is.null(tmp) && !is.null(tmp[["all"]])) {
      # a struct-like definition with slots.
     tmp = tmp[["all"]]

     slotTypes = xmlApply(tmp, function(x) {
                                  tt = xmlGetAttr(x, "type")
                                   getSOAPType(tt, types) #XXX , namespaceDefs)
                               })
     names(slotTypes) = xmlSApply(tmp, function(x) xmlGetAttr(x, "name"))
     def = ClassDef(name, slotTypes)
  } else if(!is.null(tmp) && !is.null(tmp[["extension"]])) {
     base = xmlGetAttr(tmp[["extension"]], "base")
       # Now get the extensions.
     def = ClassDef(name, list(), new("ExtendedClassDefinition"))
     def@base = base
     tmp = tmp[[1]]
     if(xmlSize(tmp) > 0 && xmlName(tmp[[1]]) == "sequence") {
        #XXXXX Taken from sequence below. Merge into 1.
      slotTypes = xmlApply(tmp[[1]], function(x) SOAPType(xmlGetAttr(x, "type"), namespaceDefs = namespaceDefs))
      names(slotTypes) = xmlSApply(tmp[[1]], function(x) xmlGetAttr(x, "name"))
      def@slotTypes = slotTypes       
     }
  } else if(!is.null(tmp)) {
    # Currently we interpret this as an Array.
    # So get the <restriction><attribute> element
    # and take the arrayType and ref attribution from that.
    restriction = tmp[["restriction"]]
    a = restriction[["attribute"]]
    def = ArrayType(xmlGetAttr(a, "arrayType"), namespaceDefs = namespaceDefs)
    def@name = name
  } else if(!is.null(tmp <- type[["all"]])) {
      # a struct-like definition with slots.
     slotTypes = xmlApply(tmp, function(x) xmlGetAttr(x, "type"))
     names(slotTypes) = xmlSApply(tmp, function(x) xmlGetAttr(x, "name"))
     def = ClassDef(name, slotTypes)
  } else if(!is.null(tmp <- type[["sequence"]])) {
        # <complexType><sequence>
      # Connect this with the code for the "all" case.
      # Need to resolve the type if it is not a primitive.
      # XXX also want the minOccurs and maxOccurs
      slotTypes = xmlApply(tmp, function(x) SOAPType(xmlGetAttr(x, "type"), namespaceDefs = namespaceDefs))
      names(slotTypes) = xmlSApply(tmp, function(x) xmlGetAttr(x, "name"))           
      def = ClassDef(name, slotTypes)
  } else if(!is.null(tmp <- type[["choice"]])) {
      slotTypes = xmlApply(tmp, function(x) xmlGetAttr(x, "type"))
      names(slotTypes) = xmlSApply(tmp, function(x) xmlGetAttr(x, "name"))               
      if(xmlSize(tmp) == 1)
        def = ClassDef(name, slotTypes[1])
      else #XXX Define UnionDef
        def = UnionDef(name, slotTypes)
  } else if(!is.null(tmp <- type[["simpleContent"]])) {
      if(xmlName(tmp[[1]]) == "extension") {
            # Assume the children are attributes for the moment.
         attrTypes = xmlApply(tmp[[1]], processWSDLType)
         names(attrTypes) = xmlSApply(tmp[[1]], xmlGetAttr(x, name))
          #XXX Add base.
         base = xmlGetAttr(tmp[[1]], "base")         
         def <- ClassDef(name, attrTypes)
      } else
          def <- NULL
  } else {
     warning("Failed to handle ", name, " of type ", xmlName(type), " & ", xmlName(type[[1]]), " in processWSDLType\n")
     def <- NULL
  }
  
  list(name = name, definition = def)
}

if(FALSE) {
# Redefined in SOAP.S.
# Which one should we use.
getSOAPType =
  #XXX FIXME - what's wrong?
function(typeName, types, namespaces)
{
  els = strsplit(typeName, ":")[[1]]

  ns = character()
  nsuri = character()
  if(length(els) > 1)  {
     # have a namespace
     idx = match(els[1], names(namespaces))
     if(!is.na(idx)) {
       nsuri = namespaces[[idx]]$uri
     }
     #XXXX if ns matches the basic one.
     new("PrimitiveSOAPType", name = els[2],
                              ns = els[1],
                              nsuri = nsuri,
                              toConverter = function(x) x,
                              fromConverter  = function(x) x
                          )
                                                
  } else
     new("BasicSOAPType", name = els[1])  
}  
}

processWSDLBindings =
  #
  # Called with a <binding>...</binding> node.
  #
function(node, doc = NULL, typeDefinitions = list())
{
 els =  node[names(node) == "operation"]

    # Get the type attribute and remove the namespace.
 type = discardNamespace(xmlGetAttr(node, "type"))

  # XPath:  portType[@name = type]
 types = doc[xmlSApply(doc, function(x) xmlName(x) == "portType" && xmlGetAttr(x, "name") == type)][[1]]

 namespaceDefinitions = doc$namespaceDefinitions
 # XXX merge with any others for this node.
 
 ops = lapply(els, processWSDLOperation, types, doc, namespaceDefinitions, typeDefinitions)
 names(ops) = sapply(els, xmlGetAttr, "name")
   
 ops  
}

discardNamespace <-
  # utility function that removes the namespace prefix from a string.
  # e.g.  foo:tagName returns tagName.
  #
  # Or  gsub("^.*:", "", type)
function(str)
{
  x = strsplit(str,":")[[1]]
  if(length(x) > 1)
    x[2]
  else
    x
}  

processWSDLOperation =
  #
  # An operation node.
  #
  #
  # types comes from the portType, not the schema.
  # The schema types are in typeDefinitions
  #
  #XXX ignores namespaces.
  #XXX should also get the documentation if there is any (e.g. see nwis).

  #XXX
function(node, types, doc, namespaceDefinitions = list(), typeDefinitions = list())
{
  name = xmlGetAttr(node, "name")

   # Get the operation element within the portType that corresponds to this operation.
  which = xmlSApply(types, function(x) xmlGetAttr(x, "name") == name)

  
  portType = types[which][[1]] # is this equivalent to types[[which]]?
  inputNode = xmlGetAttr(portType[["input"]], "message")
  input = discardNamespace(inputNode)

  
  msgs = doc[xmlSApply(doc, function(x) xmlName(x) == "message" && xmlGetAttr(x, "name") == input)]
  msg = msgs[[1]]

  args = xmlApply(msg, function(x) {
                            el =  xmlGetAttr(x, "type")
                            if(is.null(el)) 
                                el =  xmlGetAttr(x, "element")
                            
                            SOAPType(el, namespaceDefs = namespaceDefinitions)
                          })
   # need to map the name for an element rather than type. Let's do this in
   # genSOAPClientInterface() for now.
  names(args) = xmlSApply(msg, function(x) {
                                     el =  xmlGetAttr(x, "element")
                                     if(is.null(el))
                                       (xmlGetAttr(x, "name"))
                                     else 
                                       resolve(el, typeDefinitions)@name
                               })

  output = xmlGetAttr(portType[["output"]], "message")
  output = discardNamespace(output)
  
  msgs = doc[xmlSApply(doc, function(x) xmlName(x) == "message" && xmlGetAttr(x, "name") == output)]
  msg = msgs[[1]]  


  if(xmlSize(msg)) {
    value = xmlGetAttr(msg[["part"]], "type")
    if(is.null(value)) 
       value = xmlGetAttr(msg[["part"]], "element")

    if(is.null(value))
     value = NULL
  } else
     value = NULL


  namespace = xmlGetAttr(node[["input"]][["body"]], "namespace",
                              xmlGetAttr(doc, "targetNamespace", ""))

  obj = new("WSDLMethod", name = name, parameters = args,
                          returnValue = SOAPType(value, namespaceDefs = namespaceDefinitions),
                          namespace = namespace,
                          action = SOAPAction(xmlGetAttr(node[["operation"]], "soapAction")))

  obj@bindingStyle = xmlGetAttr(node[["operation"]], "style", as.character(NA))
  obj@use = sapply(c("input", "output"), 
                        function(x)
                           xmlGetAttr(node[[x]][["body"]], "use", as.charcter(NA)))

  
   # Deal with case of enclosed within <annotation>
  doc = portType[["documentation"]]
  if(is.null(doc) && !is.null(doc[["annotation"]]))
     doc = doc[["annotation"]][["documentation"]]


  if(!is.null(doc))
    obj@documentation = xmlValue(portType[["documentation"]])

  obj
}  


# Resolve finds the definition and its contents within a context
# It can resolve at one level or all the way down.
# The definitions here are typically SOAP types.

setGeneric("resolve", function(obj, context, recursive = TRUE) standardGeneric("resolve"))

setMethod("resolve", c("BasicSOAPType", "list"),
           function(obj, context, recursive = TRUE) {
             obj
           })

setMethod("resolve", c("SOAPVoidType"),
           function(obj, context, recursive = TRUE) {
             obj
           })

setMethod("resolve", c("SOAPTypeReference", "list"),
           function(obj, context, recursive = TRUE) {
             resolve(obj@name, context, recursive)
           })

setMethod("resolve", c("ArrayType", "list"),
           function(obj, context, recursive = TRUE) {
             if(recursive && (is.null(obj@elType) || length(obj@elType@name) == 0)) {
               name = obj@elementType
               obj@elType = resolve(name, context, recursive)
             }
             obj
           })

setMethod("resolve", c("character", "list"),
           function(obj, context, recursive = TRUE) {

               els = strsplit(gsub("\\\[\\\]", "", obj), ":")[[1]]
               if(length(els) == 2) {
                 if(els[1] == "xsd") {
                   return(SOAPType(els[2], els[1]))
                 }
                 els = els[2]
               } 

             
               el = match(els[1], names(context))
               if(is.na(el)) {
                  el = match(els[1], gsub(".*:", "", names(context)))
               }
               if(is.na(el)) 
                 stop("Cannot resolve SOAPTypeReference ", els[1], " in list")

               ans = context[[el]]$definition

               if(recursive && !is(ans, "TerminalSOAPType"))
                 ans = resolve(ans, context, recursive)

               ans
           })

ArrayType =
function(elementType, namespaceDefs = list(), obj = new("ArrayType"))
{
  return(parseArrayType(elementType, namespaces = namespaceDefs))
}  

ClassDef =
function(name, slotTypes, obj = new("ClassDefinition"))
{
  obj@name = name
  obj@slotTypes = slotTypes

  obj
}

